import random
from math import copysign, frexp
from six.moves import xrange
from .arbitrary import Arbitrary, with_samples
from .utils import filter_possible, until_possible, unique, product
from .primitives import elements


__all__ = ['int_', 'float_']


@Arbitrary.set_for(int)
@with_samples
class int_(Arbitrary):
    """ Uniformly distributed integer from the specified range. """

    def __init__(self, low=-1 << 32, high=(1 << 32) - 1, uniform=False):
        self.low = low
        self.high = high
        self.uniform = uniform

        if not uniform:
            self.signs = list({int(copysign(1, x)) for x in [low, high]})
            maxabs = max(abs(low), abs(high))
            minabs = min(abs(low), abs(high)) if len(self.signs) == 1 else 0
            self.minlog2 = minabs.bit_length()
            self.maxlog2 = maxabs.bit_length()

    def could_generate(self, x):
        return self.low <= x <= self.high

    @until_possible
    def next_random(self):
        if self.uniform:
            return random.randint(self.low, self.high)
        else:
            log2 = random.randint(self.minlog2, self.maxlog2)
            mn = int(2 ** (log2 - 1))
            mx = 2 ** log2
            rnd = random.randint(mn, mx - 1)
            sign = random.choice(self.signs)
            rnd *= sign
            return rnd

    @unique
    @filter_possible
    def gen_serial(self):
        yield 0
        yield 1
        yield self.low
        yield self.high
        if self.low < 0 < self.high:
            for i in xrange(0, max(abs(self.low), abs(self.high))):
                yield i
                yield -i
        else:
            # sign(low) = sign(high)
            if abs(self.low) < abs(self.high):
                # |low| < |high| thus 0 < low < high
                for i in xrange(self.low, self.high):
                    yield i
            else:
                # |high| < |low| thus low < high < 0
                for i in xrange(self.high, self.low, -1):
                    yield i


    @unique
    @filter_possible
    def shrink(self, x):
        xsign = int(copysign(1, x))
        xabs = abs(x)
        yield xsign * (xabs // 2)
        yield xsign * (xabs // 10 ** 5 * 10 ** 5)
        yield xsign * (xabs // 10 ** 2 * 10 ** 2)
        if x < 0:
            yield xabs


@unique
@filter_possible
def float_shrink(self, x):
    if x < 0:
        yield -x

    if abs(x) > 2:
        yield x / 2
    if abs(x) < 0.5:
        yield x * 2

    yield round(x)
    yield round(x, 1)
    yield round(x, 2)
    yield round(x, 3)
    yield round(x, 4)


@Arbitrary.set_for(float)
@with_samples
class float_(Arbitrary):
    """ Float with significand, exponent and sign taken according to the parameters. """

    def __init__(self, low=-100, high=100, uniform=False):
        self.low = float(low)
        self.high = float(high)
        self.uniform = uniform

        signs = list({int(copysign(1, x)) for x in [low, high]})
        self.sign = elements(*signs)

        _, max_exp = frexp(max(abs(low), abs(high)))
        assert isinstance(signs, list)
        _, min_exp = frexp(min(abs(low), abs(high))) \
                         if len(signs) == 1 else None, max_exp - 40
        self.exp = int_(min_exp, max_exp)

        self.sig = int_(0, 1 << 53, uniform=True)


    def could_generate(self, x):
        return self.low <= x <= self.high

    @until_possible
    def next_random(self):
        if self.uniform:
            return random.uniform(self.low, self.high)
        else:
            sign = self.sign.next_random()
            sig = self.sig.next_random()
            exp = self.exp.next_random()
            sig_p2 = 2 ** sig.bit_length()
            sig = 1.0 * sig / sig_p2
            return sign * sig * 2.0 ** exp

    @unique
    @filter_possible
    def gen_serial(self):
        yield 0
        yield self.low
        yield self.high

        for sign, sig, exp in product(self.sign.gen_serial(),
                                      self.sig.gen_serial(),
                                      self.exp.gen_serial()):
            sig_p2 = 2 ** sig.bit_length()
            sig = 1.0 * sig / sig_p2
            yield sign * sig * 2.0 ** exp

    shrink = float_shrink



"""Contains :class:`Arbitrary` class and the machinery it needs to work.
You shouldn't need anything other than this class from here.
"""
from abc import ABCMeta, abstractmethod
import re
import inspect
import six
from itertools import islice, count
from warnings import warn
from .utils import until_possible, filter_possible

registry = {}


class Arbitrary(six.with_metaclass(ABCMeta)):
    """Base class for all arbitraries, both built-in and custom.

    To create a custom arbitrary, you need to subclass from :class:`Arbitrary`
    and override (some or all) instance methods listed below."""

    @abstractmethod
    def next_random(self):
        """Get next random value."""
        raise StopIteration

    @abstractmethod
    def gen_serial(self):
        """Generate or return a sequence (finite or infinite) of serial values.
        """
        return []

    @abstractmethod
    def could_generate(self, x):
        """Check if x could've been generated by self.

        Required iff you use decorators like :func:`.filter_possible` or
        :func:`.until_possible`."""
        raise NotImplementedError

    @abstractmethod
    def shrink(self, x):
        """Get possible shrinkings of ``x``.

        :param x: Value previously generated by this arbitrary.
        """
        return []

    def gen_random(self):
        while True:
            yield self.next_random()

    @classmethod
    def samples_docstring_part(cls):
        try:
            args_formatted = cls.__init__.args_for_sample['def']
        except AttributeError:
            args_formatted = ''

        def append(line, indent=' ' * 4):
            res.append(indent + line)

        res = []
        append('', indent='')
        append('', indent='')
        append('.. rst-class:: html-toggle', indent='')
        append('', indent='')
        append('**Sample values**', indent='')
        append('', indent='')
        append('.. ipython::', indent='')
        append('')
        append('>>> arb = %s(%s)' % (cls.__name__, args_formatted))
        append('')
        append('@notest')
        append('>>> list(islice(arb.gen_random(), 20))')
        append('')
        append('@notest')
        append('>>> list(islice(arb.gen_serial(), 50))')
        append('')
        return '\n'.join(res)

    @staticmethod
    def args_for_sample(*args, **kwargs):
        def wrapper(method):
            method_src = inspect.getsource(method)
            m = re.match(r'\s*@Arbitrary.args_for_sample\((.*)\)', method_src)
            argsdef = m.group(1)
            method.args_for_sample = {
                'args': args,
                'kwargs': kwargs,
                'def': argsdef
            }
            return method

        return wrapper

    @staticmethod
    def set_for(type, arb=None):
        def wrapper(arb):
            if type in registry:
                warn('Arbitrary class already set for %s' % type)
            registry[type] = arb
            return arb

        if arb is None:
            return wrapper
        else:
            return wrapper(arb)

    @staticmethod
    def get_all():
        return registry[None]


def with_samples(arb_cls):
    arb_cls.__doc__ += arb_cls.samples_docstring_part()
    return arb_cls


spec_processors = []


def get_arbitrary(spec):
    if isinstance(spec, Arbitrary):
        return spec
    elif isinstance(spec, type):
        if issubclass(spec, Arbitrary):
            return spec()
        try:
            return registry[spec]()
        except KeyError:
            raise LookupError('No Arbitrary class is set for %s' % spec)

    spec_processed = [proc(spec) for proc in spec_processors]
    spec_processed = [pr for pr in spec_processed if pr is not None]
    if not spec_processed:
        raise LookupError('No match found for specification: %r' % spec)
    elif len(spec_processed) > 1:
        raise LookupError('Ambigous specification: %r' % spec)
    else:
        return spec_processed[0]


def add_spec_processor(processor):
    spec_processors.append(processor)
    return processor


class MappedArbitrary(Arbitrary):
    @abstractmethod
    def pack(self, x):
        pass

    @abstractmethod
    def unpack(self, x):
        pass

    def next_random(self):
        return self.pack(self.inner_arb.next_random())

    def gen_serial(self):
        for el in self.inner_arb.gen_serial():
            yield self.pack(el)

    def could_generate(self, x):
        return self.inner_arb.could_generate(self.unpack(x))

    def shrink(self, x):
        return map(self.pack, self.inner_arb.shrink(self.unpack(x)))


class FilteredArbitrary(Arbitrary):
    @abstractmethod
    def predicate(self, x):
        pass

    def could_generate(self, x):
        return self.inner_arb.could_generate(x) and \
               self.predicate(x)

    @until_possible
    def next_random(self):
        return self.inner_arb.next_random()

    @filter_possible
    def gen_serial(self):
        return self.inner_arb.gen_serial()

    @filter_possible
    def shrink(self, x):
        return self.inner_arb.shrink(x)




from collections import namedtuple
from itertools import islice
from .arbitrary import Arbitrary, MappedArbitrary, FilteredArbitrary, \
    get_arbitrary, with_samples, add_spec_processor
from .utils import filter_possible, unique, product
from .numbers import int_
from .primitives import char, letter

__all__ = ['tuple_', 'namedtuple_', 'dict_fixkeys',
           'list_', 'list_ordered', 'list_unique',
           'set_', 'dict_',
           'bytearray_', 'bytes_',
           'str_', 'str_letters']


@Arbitrary.set_for(tuple)
@with_samples
class tuple_(Arbitrary):
    """ Tuple with specified elements. """

    @Arbitrary.args_for_sample((int_, letter))
    def __init__(self, arbitraries):
        self.arbitraries = list(map(get_arbitrary, arbitraries))

    def could_generate(self, x):
        return all(arb.could_generate(y)
                   for arb, y in zip(self.arbitraries, x))

    def next_random(self):
        return tuple(arb.next_random() for arb in self.arbitraries)

    def gen_serial(self):
        return product(*[arb.gen_serial()
                         for arb in self.arbitraries])

    def shrink(self, x):
        for i, (arb, y) in enumerate(zip(self.arbitraries, x)):
            for shr_y in arb.shrink(y):
                yield x[:i] + (shr_y,) + x[i + 1:]


@add_spec_processor
def tuple_proc(spec):
    if isinstance(spec, tuple):
        return tuple_(map(get_arbitrary, spec))


@Arbitrary.set_for(namedtuple)
class namedtuple_(MappedArbitrary):
    def __init__(self, typename, fields):
        self.tupletype = namedtuple(typename, [name for name, _ in fields])
        self.inner_arb = tuple_([value for _, value in fields])

    def pack(self, tup):
        return self.tupletype._make(tup)

    def unpack(self, ntup):
        return tuple(ntup)


class dict_fixkeys(MappedArbitrary):
    def __init__(self, items_lst=None, **items_dct):
        self.inner_arb = namedtuple_('DictItems',
                                     items_lst or items_dct.items())

    def pack(self, ntup):
        return dict(ntup._asdict())

    def unpack(self, dct):
        ttype = self.inner_arb.tupletype
        return ttype._make([dct[name] for name in ttype._fields])


@Arbitrary.set_for(list)
@with_samples
class list_(Arbitrary):
    """ List with configurable length and elements. """

    @Arbitrary.args_for_sample(length=int_(0, 10), elements=int_(-100, 100))
    def __init__(self, length=int_(0, 20), elements=int_):
        self.length = get_arbitrary(length)
        self.elements = get_arbitrary(elements)

    def could_generate(self, x):
        return self.length.could_generate(len(x)) and \
               all(self.elements.could_generate(el) for el in x)

    def next_random(self):
        return [self.elements.next_random()
                for _ in range(self.length.next_random())]

    @unique
    @filter_possible
    def gen_serial(self):
        def gen_for_depth(depth):
            for l in self.length.gen_serial():
                if not (depth >= l or depth == 0):
                    continue

                if l == 0:
                    yield []
                else:
                    el_amount = depth + 1
                    elems = list(islice(self.elements.gen_serial(), el_amount))
                    for res in product(elems, repeat=l):
                        yield list(res)

        for depth in range(50):
            for item in gen_for_depth(depth):
                yield item

    @filter_possible
    def shrink(self, x):
        l = len(x)
        if l > 1:
            yield x[:-1]
            yield x[1:]

            yield x[:l // 3]
            yield x[l // 3:2 * l // 3]
            yield x[2 * l // 3:]

            for i, y in enumerate(x):
                for shr_y in self.elements.shrink(y):
                    yield x[:i] + [shr_y] + x[i + 1:]
        elif l == 1:
            for el in self.elements.shrink(x[0]):
                yield [el]


@add_spec_processor
def list_proc(spec):
    if isinstance(spec, list) and len(spec) == 1:
        return list_(elements=get_arbitrary(spec[0]))


@with_samples
class list_ordered(MappedArbitrary):
    """ List where elements are in order.

    .. note::

        Non-unique serially-generated values and shrink results are possible.
    """

    @Arbitrary.args_for_sample(length=int_(0, 10), elements=int_(-100, 100))
    def __init__(self, length=int_(0, 20), elements=int_,
                 key=lambda x: x, reverse=False):
        self.inner_arb = list_(length=length, elements=elements)
        self.key = key
        self.reverse = reverse

    def pack(self, lst):
        return list(sorted(lst, key=self.key, reverse=self.reverse))

    def unpack(self, lst):
        return lst


@with_samples
class list_unique(FilteredArbitrary):
    """ List with unique elements.

    .. warning::

        Implementation just keeps generating random lists until one with
        unique elements is found, so for some parameters it can be very slow.
        Also, upper length bound is never reached in some cases.
    """

    @Arbitrary.args_for_sample(length=int_(0, 10), elements=int_(-100, 100))
    def __init__(self, length=int_(0, 20), elements=int_):
        self.inner_arb = list_(length=length, elements=elements)

    def predicate(self, lst):
        return len(set(lst)) == len(lst)


@Arbitrary.set_for(set)
@with_samples
class set_(MappedArbitrary):
    """ Set of unique elements.

    .. warning::

        Implementation just keeps generating random lists until one with
        unique elements is found, so for some parameters it can be very slow.
        Also, upper length bound is never reached in some cases.

    .. note::

        Non-unique serially-generated values and shrink results are possible.
    """

    def __init__(self, length=int_(0, 20), elements=int_):
        self.inner_arb = list_unique(length=length, elements=elements)

    def pack(self, lst):
        return set(lst)

    def unpack(self, set_val):
        return list(set_val)


@Arbitrary.set_for(dict)
@with_samples
class dict_(MappedArbitrary):
    """ Dictionary with specified items (key-value pairs) arbitrary.

    .. note::

        Non-unique serially-generated values and shrink results are possible.
    """

    def __init__(self,
                 length=int_(0, 20),
                 items=None,
                 keys=int_, values=int_):
        if items is None:
            items = tuple_((keys, values))
        self.inner_arb = list_(length=length, elements=items)

    def pack(self, lst):
        return dict(lst)

    def unpack(self, dct):
        return list(dct.items())

    @unique
    def gen_serial(self):
        return super(dict_, self).gen_serial()


@Arbitrary.set_for(bytearray)
@with_samples
class bytearray_(MappedArbitrary):
    """ Mutable bytearray with configurable length and elements. """

    def __init__(self, length=int_(0, 20), elements=int_(0, 255)):
        self.inner_arb = list_(length=length, elements=elements)

    def pack(self, lst):
        return bytearray(lst)

    def unpack(self, barr):
        return list(barr)


@with_samples
class bytes_(MappedArbitrary):
    """ Immutable bytearray with configurable length and elements.

    .. note::

        In Python 2 it's the same as :class:`str_`.
    """

    def __init__(self, length=int_(0, 20), elements=int_(0, 255)):
        self.inner_arb = bytearray_(length=length, elements=elements)

    def pack(self, barr):
        return bytes(barr)

    def unpack(self, bs):
        return bytearray(bs)


if bytes is not str:
    Arbitrary.set_for(bytes, bytes_)


@Arbitrary.set_for(str)
@with_samples
class str_(MappedArbitrary):
    """ String with configurable length and characters. """

    def __init__(self, length=int_(0, 20), chars=char):
        self.inner_arb = list_(length=length, elements=chars)

    def pack(self, lst):
        return ''.join(lst)

    def unpack(self, s):
        return list(s)


@with_samples
class str_letters(str_):
    """ String consisting of letters only. """

    def __init__(self, length=int_(0, 20)):
        str_.__init__(self, length=length, chars=letter)



from collections import OrderedDict
from functools import wraps
import inspect
from .checker import Checker
from .utils import optional_args


@optional_args
def qc(ntests=None, nshrinks=None, nassume=None):
    """ qc(ntests=None, nshrinks=None, nassume=None)
    Decorator for a test function.

    For usage examples see :ref:`creating-tests`.

    :param ntests: number of tests
    """

    frames = inspect.stack()
    func_is_method = any(fr[4][0].strip().startswith('class ')
                         for fr in frames
                         if fr and fr[4])

    def wrapper(func):
        argspec = inspect.getargspec(func)
        argnames = list(argspec.args or [])
        defaults = list(argspec.defaults or [])
        defaults = [None] * (len(argnames) - len(defaults)) + defaults
        args = OrderedDict(zip(argnames, defaults))

        if any(args[a] is None for a in args):
            raise TypeError('unbound variables: %s' %
                            ', '.join(a for a in args if args[a] is None))

        checker = Checker(func, args, ntests, nshrinks, nassume, ignore_return=True)

        def wrapped(*args):
            if not (len(args) == 0 or len(args) == 1 and func_is_method):
                raise ValueError('not expecting any arguments')

            result = checker.for_all()
            if result.failed:
                result.reraise()

        wraps(func)(wrapped)
        wrapped.__test__ = True
        return wrapped

    return wrapper



""" Contain utility functions and decorators useful when defining custom
:class:`.Arbitrary` instances. Usually they haven't much use otherwise.
"""

from itertools import cycle, islice, count, product as product_
import random
from decorator import decorator
from six import PY3
from six.moves import zip_longest
from pyquchk.lazylist import LazyList

__all__ = ['roundrobin', 'interleave', 'product',
           'choice_weighted',
           'filter_possible', 'until_possible', 'unique']


@decorator
def filter_possible(method, self, *args, **kwargs):
    """ Decorator, when applied to a generator method (e.g. :meth:`.gen_serial`)
    filters generated values to keep only those for which :meth:`.could_generate`
    returns ``True``.
    """
    fails_cnt = 0
    for item in method(self, *args, **kwargs):
        if self.could_generate(item):
            yield item
            fails_cnt = 0
        else:
            fails_cnt += 1
            if fails_cnt > 10000:
                raise RuntimeError('Unable to satisfy could_generate after 10000 items from gen_serial in a %s instance.' % type(self).__name__)


@decorator
def until_possible(method, self, *args, **kwargs):
    """ Decorator, when applied to a method returning different values each time
    (e.g. :meth:`.next_random`) runs this method multiple times until a value
    for which :meth:`.could_generate` returns ``True`` is returned, and finally
    returns this value.
    """
    fails_cnt = 0
    func = lambda: method(self, *args, **kwargs)
    for item in iter(func, object()):
        if self.could_generate(item):
            return item
        else:
            fails_cnt += 1
            if fails_cnt > 10000:
                raise RuntimeError('Unable to satisfy could_generate after 10000 calls of next_random in a %s instance.' % type(self).__name__)


@decorator
def unique(method, self, *args):
    """ Decorator, when applied to a generator method (e.g. :meth:`.gen_serial`,
    :meth:`.shrink`) filters generated values to keep only unique values.
    For the :meth:`.shrink` method the value passed to the method is also
    considered.
    """
    seen = set()
    if 'shrink' in method.__name__:
        seen.add(args[0])

    for item in method(self, *args):
        if isinstance(item, list):
            s_item = tuple(item)
        elif isinstance(item, dict):
            s_item = tuple(item.items())
        else:
            s_item = item

        if s_item not in seen:
            yield item
            seen.add(s_item)


def roundrobin(iterables):
    """ Yield items from the given iterables in a round-robin fashion.

    What it does exactly is better explained with a simple example:

    .. ipython::

        >>> list(roundrobin(['ABC', 'D', 'EF']))
        ['A', 'D', 'E', 'B', 'F', 'C']

    :param iterables: List of finite or infinite iterables.
    """
    # Recipe credited to George Sakkis
    pending = len(iterables)
    iterators = (iter(it) for it in iterables)
    nexts = cycle(it.__next__ if PY3 else it.next for it in iterators)
    while pending:
        try:
            for next_f in nexts:
                yield next_f()
        except StopIteration:
            pending -= 1
            nexts = cycle(islice(nexts, pending))


def interleave(xs, ys):
    """
    >>> list(interleave([], []))
    []
    >>> list(interleave([1], []))
    [1]
    >>> list(interleave([], ['a']))
    ['a']
    >>> list(interleave([1], ['a']))
    [1, 'a']
    >>> list(interleave([1, 2], ['a']))
    [1, 'a', 2]
    >>> list(interleave([1], ['a', 'b']))
    [1, 'a', 'b']
    >>> list(interleave([1, 2], ['a', 'b']))
    [1, 'a', 2, 'b']
    """
    for tup in zip_longest(xs, ys):
        for x in tup:
            if x is not None:
                yield x


def product(*iterables, **kwargs):
    """ product(*iterable, product=1)
    Cartesian product of all the iterables.

    Difference compared to the :func:`itertools.product` is that tuples are yield
    in the order suitable for infinite iterables (see examples below).

    Corner cases:

    .. ipython::

        >>> list(product([], []))
        []

        >>> list(product([1, 2, 3], []))
        []

        >>> list(product([], [1, 2, 3]))
        []

    2 iterables:

    .. ipython::

        >>> list(product([1], [1]))
        [(1, 1)]

        >>> list(product([1, 2], [1]))
        [(1, 1), (2, 1)]

        >>> list(product([1, 2], [1, 2]))
        [(1, 1), (1, 2), (2, 1), (2, 2)]

        >>> list(product([1, 2, 3], [1, 2, 3]))
        [(1, 1), (1, 2), (2, 1), (2, 2), (1, 3), (2, 3), (3, 1), (3, 2), (3, 3)]

        >>> list(product([1, 2, 3], [1, 2, 3])) == list(product([1, 2, 3], repeat=2))
        True

        >>> list(islice(product(count(1), count(1)), 1000)) == list(islice(product(count(1), repeat=2), 1000))
        True

        >>> infNN = product(count(1), count(1))

        >>> list(islice(infNN, 20))
        [(1, 1), (1, 2), (2, 1), (2, 2), (1, 3), (2, 3), (3, 1), (3, 2), (3, 3), (1, 4), (2, 4), (3, 4), (4, 1), (4, 2), (4, 3), (4, 4), (1, 5), (2, 5), (3, 5), (4, 5)]

        >>> lstNN = list(islice(infNN, 10000))

        >>> all(max(x) <= max(y) for x, y in zip(lstNN, lstNN[1:]))
        True

    3 iterables:

    .. ipython::

        >>> list(product([1, 2], [1, 2], [1, 2]))
        [(1, 1, 1), (1, 1, 2), (1, 2, 1), (1, 2, 2), (2, 1, 1), (2, 1, 2), (2, 2, 1), (2, 2, 2)]

        >>> len(list(product([1, 2, 3], [1, 2, 3], [1, 2, 3])))
        27

        >>> infNNN = product(count(1), count(1), count(1))
        >>> list(islice(infNNN, 20))
        [(1, 1, 1), (1, 1, 2), (1, 2, 1), (1, 2, 2), (2, 1, 1), (2, 1, 2), (2, 2, 1), (2, 2, 2), (1, 1, 3), (1, 2, 3), (2, 1, 3), (2, 2, 3), (1, 3, 1), (1, 3, 2), (1, 3, 3), (2, 3, 1), (2, 3, 2), (2, 3, 3), (3, 1, 1), (3, 1, 2)]

        >>> lstNNN = list(islice(infNNN, 10000))
        >>> all(max(x) <= max(y) for x, y in zip(lstNNN, lstNNN[1:]))
        True
    """
    iterables = list(map(LazyList, iterables)) * kwargs.get('repeat', 1)

    for i in count():
        yielded_any = False

        for main_j, main_it in reversed(list(enumerate(iterables))):
            try:
                x = main_it[i]
            except IndexError:
                continue

            prod = product_(*[it[:i + 1 if j > main_j else i] for j, it in
                              enumerate(iterables) if j != main_j])

            for tup in prod:
                yield tup[:main_j] + (x,) + tup[main_j:]
                yielded_any = True

        if not yielded_any:
            break


def choice_weighted(elements):
    """ Like :func:`random.choice`, but the probability an item is chosen is
    proportional to its weight.

    For convenience, this function is available as ``random.choice_weighted``
    if you imported :mod:`.utils`.

    :param elements: list of tuples ``(item, weight)``, weight is any numeric value
    """
    rnd = random.random() * sum(w for _, w in elements)
    for el, w in elements:
        rnd -= w
        if rnd < 0:
            return el


random.choice_weighted = choice_weighted



""" Contains all the built-in :class:`.Arbitrary` instances. They all can be
imported with ``from pyquchk.arbitraries import ...``, no need to specify
exact modules.
"""
from .primitives import *
from .numbers import *
from .sequences import *



import sys
from warnings import warn
import six
from .arbitraries.arbitrary import get_arbitrary
from .arbitraries.sequences import dict_fixkeys


__all__ = ['for_all', 'Result', 'ReturnedFalse', 'CheckError',
           'AssumptionError', 'assume']


class AssumptionError(Exception):
    pass


def assume(condition):
    if not condition:
        raise AssumptionError


class CheckError(Exception):
    """ Raised when a test which uses :func:`.qc` decorator fails.

    It contains the original test data for which the test has failed, and
    the exception raised by the tested function.
    Actually, a :exc:`CheckError` instance is always also an instance of the
    exception raised by the tested function, so you can except it as usually.
    """

    #: Data which caused the function to raise an exception
    test_data = None
    #: Exception raised by the function, or :exc:`ReturnedFalse` if function
    #: returned ``False``
    cause = None


def get_CheckError(data, noshrink_data, cause):
    class CheckError_(CheckError, type(cause)):
        def __init__(self, data, noshrink_data, cause):
            self.test_data = data
            self.noshrink_data = noshrink_data
            self.cause = cause

        def __str__(self):
            msg = "test failed"
            if self.cause is not None:
                msg += " due to %s" % type(self.cause).__name__
                cause = str(self.cause)
                if cause:
                    msg += ": " + cause

            res = [msg]
            res.append('')
            res.append("Failure encountered for data:")
            res += ["  %s = %r" % i for i in sorted(self.test_data.items())]
            res.append('')
            res.append("Before shrinking:")
            res += ["  %s = %r" % i for i in sorted(self.noshrink_data.items())]
            return '\n'.join(res)

    CheckError_.__name__ = 'CheckError'
    return CheckError_(data, noshrink_data, cause)


class ReturnedFalse(Exception):
    """ Automatically raised when the tested function returns a value which
    evaluates to a :mod:`bool` ``False``.
    """

    def __init__(self, retval):
        #: Exact value returned by the function (evaluates to a ``False`` :mod:`bool` value)
        self.retval = retval

    def __str__(self):
        return 'function returned %r' % self.retval


class Result(object):
    """ Represent the result of performed testing (returned by :func:`for_all`).
    """

    def __init__(self,
                 verdict, passed, datas,
                 data=None, noshrink_data=None, exception=None, traceback=None):
        #: ``'OK'`` if passed and ``'FAIL'`` if not
        self.verdict = verdict
        #: ``True`` if passed and ``False`` if not (opposite to :attr:`failed`)
        self.passed = passed
        #: ``False`` if passed and ``True`` if not (opposite to :attr:`passed`)
        self.failed = not passed
        #: List of all the data values generated during the test run
        self.datas = datas
        #: Data on which the tested function failed (as a :mod:`dict`),
        #: or ``None`` if passed
        self.data = data
        #: Same as :attr:`data`, but before shrinking.
        self.noshrink_data = noshrink_data
        #: Exception raised by the function, or :exc:`ReturnedFalse` if function
        #: returned ``False``, or ``None`` if passed
        self.exception = exception
        #: Exception traceback, or ``None`` if passed
        self.traceback = traceback

    @classmethod
    def ok(cls, datas):
        return Result('OK', True, datas)

    @classmethod
    def fail(cls, datas, data, noshrink_data, exception, traceback):
        return Result('FAIL', False,
                      datas, data, noshrink_data,
                      exception, traceback)

    def reraise(self):
        exception = get_CheckError(self.data, self.noshrink_data,
                                   self.exception)
        six.reraise(type(exception), exception, self.traceback)

    def __bool__(self):
        """ Return :attr:`passed`, used to interprete as a :mod:`bool` value.
        """
        return self.passed

    __nonzero__ = __bool__


class Checker(object):
    def __init__(self, func, args, ntests, nshrinks, nassume, ignore_return=False):
        self.func = func
        self.args = args
        self.ntests = ntests or 500
        self.nshrinks = nshrinks or 1000
        self.nassume = nassume or 1000
        self.ignore_return = ignore_return

        arg_arbs = {argname: get_arbitrary(self.args[argname])
                    for argname in self.args}
        self.arb = dict_fixkeys(**arg_arbs)


    def eval_func(self, data):
        func_res = self.func(**data)
        if not self.ignore_return and not func_res:
            raise ReturnedFalse(func_res)


    def full_shrink(self, kwargs):
        n = 0
        while n < self.nshrinks:
            for shr_kwargs in self.arb.shrink(kwargs):
                n += 1
                try:
                    self.eval_func(shr_kwargs)
                except AssumptionError:
                    continue
                except Exception as e:
                    kwargs = shr_kwargs
                    break
                if n > self.nshrinks:
                    break
            else:
                break
        else:
            warn('Shrink did not stop after %d iterations.' % self.nshrinks)
        return kwargs


    def for_all(self):
        gen_serial = self.arb.gen_serial()

        def generate(n):
            if n < self.ntests / 2:
                try:
                    return next(gen_serial)
                except StopIteration:
                    return self.arb.next_random()
            else:
                return self.arb.next_random()

        datas = []
        n = 0
        assume_fail_cnt = 0
        while n < self.ntests:
            kwargs = generate(n)
            datas.append(kwargs)
            try:
                self.eval_func(kwargs)
            except AssumptionError:
                del datas[-1]
                assume_fail_cnt += 1
                if assume_fail_cnt > self.nassume:
                    raise RuntimeError(
                        'Could not satisfy assume() after %d of test cases.' % self.nassume)
                continue
            except Exception as e:
                shrinked_kwargs = self.full_shrink(kwargs)
                try:
                    self.eval_func(shrinked_kwargs)
                except Exception as e:
                    _, _, traceback = sys.exc_info()
                    return Result.fail(datas, shrinked_kwargs, kwargs, e,
                                       traceback)
            assume_fail_cnt = 0
            n += 1

        return Result.ok(datas)


def for_all(func, **kwargs):
    """ for_all(func, ntests=None, nshrinks=None, nassume=None)
    Test the function ``func`` on ``_ntests`` tests.

    For usage examples see :ref:`property-checking`.

    :param kwargs: function arguments specification
    :returns: information about the tests run
    :rtype: :class:`Result`
    """
    ntests = kwargs.get('ntests', None)
    nshrinks = kwargs.get('nshrinks', None)
    nassume = kwargs.get('nassume', None)
    return Checker(func, kwargs, ntests, nshrinks, nassume).for_all()



from pyquchk import *
from pyquchk.arbitraries import *
from pytest import *


def test_simple_true():
    assert for_all(lambda x: x * 2 / 2 == x, x=int_)


def test_simple_false():
    assert not for_all(lambda x: x * 3 / 2 == x, x=int_)


def test_float_add_associativity():
    assert not for_all(lambda a, b, c: a + (b + c) == (a + b) + c,
                       a=float, b=float, c=float)
    assert not for_all(lambda a, b, c: a + (b + c) == (a + b) + c,
                       a=float_(1, 1.1), b=float_(1, 1.1), c=float_(1, 1.1))
    assert not for_all(lambda lst: sum(lst) == sum(reversed(lst)),
                       lst=list_(length=int_(1, 20), elements=float_()))


def test_assume():
    assert not for_all(
        lambda x, y: len(x) == len(y),
        x=list, y=list)
    res = for_all(
        lambda x, y: assume(len(x) == len(y)) is None and len(x) == len(y),
        x=list, y=list)
    assert res
    assert len(res.datas) == 500
    with raises(RuntimeError):
        for_all(lambda x: assume(False), x=int)



from functools import wraps
import inspect


def optional_args(decor):
    """Decorator for decorators (sic) that are intended to take
    optional arguments.

    It supports decorators written both as classes or functions,
    as long as they are "doubly-callable".
    For classes, this means implementing ``__call__``, while
    functions must return a function that returns a function
    that accepts a function... which is obvious, of course.
    """

    @wraps(decor)
    def wrapped(*args, **kwargs):
        one_arg = len(args) == 1 and not kwargs
        if one_arg and inspect.isfunction(args[0]):
            return decor()(args[0])
        else:
            return decor(*args, **kwargs)

    return wrapped




from itertools import count


class LazyList(object):
    """A Sequence whose values are computed lazily by an iterator.

    .. ipython::

        >>> it = count()

        >>> ll = LazyList(it)

        >>> list(ll[:10])
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

        >>> list(ll[:10])
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

        >>> list(ll[:10])
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

        >>> ll[123]
        123

        >>> list(ll[1:10:2])
        [1, 3, 5, 7, 9]
    """

    def __init__(self, iterable):
        self._exhausted = False
        self._iterator = iter(iterable)
        self._data = []

    def __len__(self):
        """Get the length of a LazyList's computed data."""
        return len(self._data)

    def __getitem__(self, i):
        """Get an item from a LazyList.
        i should be a positive integer or a slice object."""
        if isinstance(i, int):
            #index has not yet been yielded by iterator (or iterator exhausted
            #before reaching that index)
            if i >= len(self):
                self.exhaust(i)
            elif i < 0:
                raise ValueError('cannot index LazyList with negative number')
            return self._data[i]

        #LazyList slices are iterators over a portion of the list.
        elif isinstance(i, slice):
            start, stop, step = i.start, i.stop, i.step
            if any(x is not None and x < 0 for x in (start, stop, step)):
                raise ValueError('cannot index or step through a LazyList with'
                                 'a negative number')
                #set start and step to their integer defaults if they are None.
            if start is None:
                start = 0
            if step is None:
                step = 1

            def LazyListIterator():
                count = start
                while stop is None or count < stop:
                    try:
                        yield self[count]
                    #slices can go out of actual index range without raising an
                    #error
                    except IndexError:
                        break
                    count += step

            return LazyListIterator()

        raise TypeError('i must be an integer or slice')

    def __iter__(self):
        """Return an iterator over each value in the sequence,
        whether it has been computed yet or not."""
        return self[:]

    def computed(self):
        """Return an iterator over the values in a LazyList that have
        already been computed."""
        return self[:len(self)]

    def exhaust(self, index=None):
        """Exhaust the iterator generating this LazyList's values.
        if index is None, this will exhaust the iterator completely.
        Otherwise, it will iterate over the iterator until either the list
        has a value for index or the iterator is exhausted.
        """
        if self._exhausted:
            return
        if index is None:
            ind_range = count(len(self))
        else:
            ind_range = range(len(self), index + 1)

        for ind in ind_range:
            try:
                self._data.append(next(self._iterator))
            except StopIteration: #iterator is fully exhausted
                self._exhausted = True
                break



from pyquchk import *
from pyquchk.arbitraries import *
from pytest import *


@qc
def test_out_of_class_definition(a=int_):
    assert type(a) == int


def test_unbounded_var():
    with raises(TypeError):
        @qc
        def unbounded_var(self, a=int_):
            pass


def test_unbounded_var_in_class():
    with raises(TypeError):
        class test_class:
            @qc
            def unbounded_var(self, a=int_):
                pass


class Test_spec_options:
    @qc
    def test_arb_noparam(a=int_, b=int_):
        assert type(a) == int
        assert type(b) == int
        assert a + b == b + a

    @qc
    def test_type_spec(a=int, b=int):
        assert type(a) == int
        assert type(b) == int
        assert a + b == b + a

    @qc
    def test_arb_withparam(a=int_(low=1000), b=int_(low=1 << 30)):
        assert type(a) == int
        assert type(b) == int
        assert a >= 1000
        assert b >= 1 << 30
        assert a + b == b + a


class Test_numbers:
    @qc
    def test_int_add_associativity(a=int, b=int, c=int):
        assert a + (b + c) == (a + b) + c


class Test_list:
    @qc
    def test_int_ops(lst=list_):
        if lst:
            avg_i = sum(lst) / len(lst)
            avg_f = 1.0 * sum(lst) / len(lst)
        assert (len(lst) == 0
                or len(set(lst)) == 1 and min(lst) == avg_i == max(lst)
                or min(lst) < avg_f < max(lst))

    @qc
    def test_shrink_props(lst=list_, n=int_(0, 5), m=int_(0, 5)):
        shrinked_n = list(list_(length=int_(n, n)).shrink(lst))
        assert all(len(l) == n for l in shrinked_n)

        shrinked_nm = list(list_(length=int_(n, m)).shrink(lst))
        assert all(n <= len(l) <= m for l in shrinked_nm)


@qc
def test_with_assume(x=list, y=list):
    assume(len(x) == len(y))
    assert len(x) == len(y)



""" Contains main functions and facilities you will need when using
:mod:`pyquchk`. They all can be imported with ``from pyquchk import ...``,
no need to specify exact modules.
"""

from .qc_decorator import qc
from .checker import *



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys


class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import tox

        errno = tox.cmdline(self.test_args)
        sys.exit(errno)


with open('README.rst') as file:
    long_description = file.read()

setup(
    name='pyquchk',
    version='0.5.0',
    packages=find_packages(),
    url='',
    license='',
    author='Alexander Plavin',
    author_email='alexander@plav.in',
    description='Easy-to-use, extensible random testing library.',
    long_description=long_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Testing',
    ],
    tests_require=['pytest'],
    install_requires=['six', 'decorator'],
    cmdclass={'test': Tox},
)



import random
from string import ascii_letters
from .arbitrary import Arbitrary, get_arbitrary, with_samples
from .utils import roundrobin, filter_possible, unique


__all__ = ['const', 'elements', 'oneof', 'bool_', 'char', 'letter']


@with_samples
class elements(Arbitrary):
    """ Choice from a list of constant values. """

    @Arbitrary.args_for_sample('a', 'b', 'c', 'd')
    def __init__(self, *el_list):
        self.el_list = el_list

    def could_generate(self, x):
        return x in self.el_list

    def next_random(self):
        return random.choice(self.el_list)

    def gen_serial(self):
        return self.el_list

    def shrink(self, x):
        for el in self.el_list:
            if el == x:
                break
            yield el


@Arbitrary.set_for(bool)
@with_samples
class bool_(elements):
    """ Any :mod:`bool` value (``False`` or ``True``). """

    def __init__(self):
        elements.__init__(self, False, True)


@with_samples
class const(elements):
    """ Any constant value. """

    @Arbitrary.args_for_sample(value='abc')
    def __init__(self, value=None):
        elements.__init__(self, value)


@with_samples
class char(elements):
    """ Any single character (:mod:`string` of length one). """

    def __init__(self, chars=list(map(chr, range(0, 255)))):
        elements.__init__(self, *chars)

    @unique
    @filter_possible
    def shrink(self, x):
        lst = ['a', 'z', 'A', 'Z', '0', '9', ' ',
               self.el_list[0], self.el_list[-1], chr(0), chr(255)]
        for c in lst:
            if c == x:
                break
            yield c


@with_samples
class letter(char):
    """ Lower and upper latin letters. """

    def __init__(self):
        char.__init__(self, ascii_letters)


@with_samples
class oneof(Arbitrary):
    """ Choice (possible weighed) from a list of Arbitraries. """

    @Arbitrary.args_for_sample(const('a'), const('b'), const('c'))
    def __init__(self, *arbitraries):
        arbitraries = [arb if isinstance(arb, tuple) else (arb, 1)
                       for arb in arbitraries]
        self.arbitraries = [(get_arbitrary(arb), w) for arb, w in arbitraries]

    def could_generate(self, x):
        return any(arb.could_generate(x) for arb, _ in self.arbitraries)

    def next_random(self):
        arb = random.choice_weighted(self.arbitraries)
        return arb.next_random()

    def gen_serial(self):
        inner_series = [arb.gen_serial() for arb, _ in self.arbitraries]
        return roundrobin(inner_series)

    def shrink(self, x):
        return []
